%%
%%  IKEv2-MAC
%%  
%%  Modelling follows the one from the AVISPA Library of Protocols.
%%


% Functors:

    % cryptographic primitives
    fun sig/2.
    fun enc/2.
    fun f/1.

    % participant names 
    fun a/0.
    fun b/0.
    fun c/0.

    % constants
    fun g/0.
    fun sec/0.

    % nonces and keys
    fun k/1.    % k(K) -- private key of X.
    fun pub/1.

    fun psk/2.  % psk(X,Y) -- a symmetric shared by X and Y.

    fun sa1/2.   % ...(X,Y) -- a nonce generated by X in the session with Y.
    fun sa2/2.  
    fun dhx/2. 
    fun dhy/2.
    fun ni/2.
    fun nr/2.

    % fun km/1.   % km(S) -- a key derived from a secret S
    % fun ke/1.   % as above
    % fun ks/1.   % as above

    fun completedA/6.
    fun completedB/6.


%%%%% AUXILIARY PREDICATES %%%%%%%%%%%%

participant(a).
participant(b).
participant(c).

%%
% The following predicates specify which session will be
% taken into consideration.

% roleA(X,Y)  ---  X as an initiator runs the protocol with Y as responder 
% roleB(X,Y)  ---  X as a responder runs the protocol with Y as initiator 

roleA(a, a, psk(a,a)).
% roleA(a, b, psk(a,b)).
% roleA(b, a, psk(a,b)).
roleA(a, c, psk(a,c)).
% roleA(b, c, psk(b,c)).

roleB(a, a, psk(a,a)).
% roleB(a, b, psk(a,b)).
% roleB(b, a, psk(a,b)).
roleB(a, c, psk(a,c)).
% roleB(b, c, psk(b,c)).

% roleA( A, B, PSK, X, NI, SA1, SA2 )
%
roleA( A, B, PSK, dhx(A,B), ni(A,B), sa1(A,B), sa2(A,B) ) :- roleA(A,B,PSK).
    
% roleB( B, A, PSK, Y, NR )
%
roleB(B, A, PSK, dhy(B,A), nr(B,A)) :- roleB(B,A,PSK).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The secret:

    query sec.

% Intruder rules

    rule:: sig(_,M) -> M.
    rule:: sig(K,_) -> pub(K).
    rule:: K, M -> sig(K,M).
    rule:: K, M -> enc(K,M).
    rule:: K, enc(K,M) -> M.
    rule:: M -> f(M).

    % rule:: X -> km(X).
    % rule:: X -> ke(X).
    % rule:: X -> ks(X).

% Initial intruder knowledge

    rule:: X             :- participant(X).
    rule:: pub(k(X))     :- participant(X).
    rule:: k(c).
    rule:: g.

    rule:: psk(c,X)        :- participant(X).
    rule:: psk(X,c)        :- participant(X).

% Protocol Rules (the main session):


    % (1) A->B
    %
    rule::
        SA1, g^X, NI 
    :-
        roleA( _A, _B, _PSK, X, NI, SA1, _SA2 ).

    % (2) B->A
    %
    rule::
        SA1, _GX, _NI
      ->
        SA1, g^Y, NR
    :-
        roleB(_B, _A, _PSK, Y, NR).


    % (3) A->B
    %
    rule::
        SA1, GY, NR
      ->
        enc(SK, (A, f((PSK, SA1, g^X, NI, NR)), SA2))
    :-
        roleA( A, _B, PSK, X, NI, SA1, SA2 ),
        SK = f((NI, NR, SA1, GY^X)).


    % (4) B->A
    %
    rule::
        SA1, GX, NI, % from the previous rule
        enc(SK, (A, f((PSK, SA1, GX, NI, NR)), SA2))
      ->
        enc(SK, (B, f((PSK, SA1, g^Y, NI, NR)), SA2)),
        completedB(B, A, GX, g^Y, GX^Y, SK)
    :-
        roleB(B, A, PSK, Y, NR),
        SK = f((NI, NR, SA1, GX^Y)).

    % (5) A completed
    %
    rule::
        NR, GY, % from the precious rule
        enc(SK, (B, f((PSK, SA1, GY, NI, NR)), SA2))
      ->
        completedA(A, B, g^X, GY, GY^X, SK)
    :-
        roleA( A, B, PSK, X, NI, SA1, SA2 ),
        SK = f((NI, NR, SA1, GY^X)).


%%%%%%%%%%%%%%%%%%%%

    rule:: SK, completedA(a,a,_,_,_,SK) -> sec.
    rule:: SK, completedB(a,a,_,_,_,SK) -> sec.

    rule:: GXY, completedA(a,a,_,_,GXY,_) -> sec.
    rule:: GXY, completedB(a,a,_,_,GXY,_) -> sec.

    rule:: g^dhx(a,a)^dhy(a,a) -> sec.

    rule:: completedA(a, a, _, g^dhy(a,c), _, _) -> sec.
    rule:: completedB(a, a, g^dhx(a,c), _, _, _) -> sec. 

    %%%
    %% should fire
    %
    % rule:: completedA(a,a,GX,GY,GXY,SK), completedB(a,a,GX,GY,GXY,SK) -> sec.
    % rule:: SK, completedA(a,c,_,_,_,SK) -> sec.
    % rule:: SK, completedB(a,c,_,_,_,SK) -> sec.
    
