%%
%%  KERBEROS
%%  
%%

%%% cryptographic primitives

    fun sig/2.  % sig(K,M) -- a non-message-recovering signature on M
    fun enc/2.  % enc(K,M) -- symmetric encryption of M with key K
    fun h/1.    % h(M) -- hash of M

%%% constants and nonces
    fun g/0.
    fun sec/0.
    fun n1/1.   % n1(C) -- a nonce generated by C
    fun n2/1.   % n2(C) -- a nonce generated by C
    fun n3/1.   % n3(C) -- a nonce generated by C
    fun nx/1.   % nx(C) -- a nonce (exponent) generated by C
    fun ny/1.   % ny(C) a nonce (exponent) generated by kas for C
    fun kT/0.   % a key shared by kas and tgs
    fun kS/0.   % a key shared by tgs and serv
    fun ak/2.   % ak(C,...) -- authentication key generated by kas for C
    fun sk/1.   % sk(C) -- service key generated by tgs for C
    fun key/1.  % key(C) -- the private key of C
    fun pub/1.  % the corresponding public key

    fun request/1.

    fun completed/2.
    fun completed_C_auth/4.
    fun completed_C_tgs/4.
    fun completed_C/3.
    fun completed_S/3.
    fun completed_TGS/3.

%%% participants and roles

    fun a/0.    % client
    fun c/0.    % client (the intruder)
    fun kas/0.  % Kerberos Authentication Server
    fun tgs/0.  % Ticket Granting Server
    fun serv/0. % Server

    client(a).
    client(c).
    honest_client(a).
    participant(a).
    participant(c).
    participant(kas).
    participant(tgs).
    participant(serv).

%%% The secret:

    query sec.

%%% Intruder rules

    rule:: sig(_,M) -> M. 
    rule:: sig(K,_) -> pub(K). 
    rule:: K, M -> sig(K,M).
    rule:: K, M -> enc(K,M).
    rule:: K, enc(K,M) -> M.
    rule:: M -> h(M).

%%% Initial intruder knowledge

    rule:: X              :- participant(X).
    rule:: pub(key(X))    :- participant(X).
    rule:: key(c).
    rule:: g.

% Protocol Rules:

    %%
    % (1) [C -> KAS]
    %
    rule::  
        sig(key(C), (h((C,tgs,n1(C))), n2(C), g^nx(C))), 
        C, tgs, n1(C)
    :- 
        honest_client(C). 

    %%
    % (2) [KAS -> C]
    %
    rule::
        sig(key(C), (h((C,tgs,N1)), N2, GX)), 
        C, tgs, N1
      -> 
        sig(key(kas), (g^ny(C), N2)),  C,  
        enc(kT, (ak(C,GX), C)),    % TGT (Ticket Granting Ticket)
        enc(GX^ny(C), (ak(C,GX), N1, tgs)),
        completed(kas, (C, ak(C,GX), GX^ny(C)))
    :-
        client(C).


    %%
    % (3) [C -> TGS]
    %
    rule::
        sig(key(kas), (GY, n2(C))),  C,  TGT,    
        enc(GY^nx(C), (AK, n1(C), tgs))
      ->
        TGT,  enc(AK, C),  serv,  n3(C),
        completed_C_auth(C, AK, TGT, GY^nx(C) )
    :-
        honest_client(C).

    %%
    % (4) [TGS -> C]
    %
    rule::
        enc(kT, (AK, C)),   % TGT 
        enc(AK, C),
        serv, N3
      ->
        C, 
        enc(kS, (sk(C),C)), % ST
        enc(AK, (sk(C),N3,serv)), 
        completed_TGS(C,AK,N3)
    :-
        % C = c.
        client(C).


    %%
    % (5) [C -> S]
    %
    rule::
      % from previous rule:
        sig(key(kas), (GY, n2(C))), 
        enc(GY^nx(C), (AK, n1(C), tgs)),
      % from this rule:
        C,  ST,  
        enc(AK, (SK,n3(C),serv))
      ->
        ST,  enc(SK, (C, request(C))),
        completed_C_tgs(C, AK, ST, SK)
    :-
        honest_client(C).


    %%
    % (6) [S -> C]
    %
    rule::
        enc(kS, (SK,C)), % ST
        enc(SK, (C, Request))
      ->
        enc(SK, Request),
        completed_S(C,SK,Request)
    :-
        client(C).


    %%
    % (7) C completed
    %
    rule::
      % from previous rule:
        sig(key(kas), (GY, n2(C))), 
        enc(GY^nx(C), (AK, n1(C), tgs)),
        enc(AK, (SK,n3(C),serv)),
      % from this rule:
        enc(SK, request(C))
      ->
        completed_C(C, AK, SK)
    :-
        honest_client(C).


%%%%%%

    %%%
    %% should not fire
    %
    
    rule:: completed_C_auth(a, ak(c,_), _, _)  -> sec.  % !!!
    
    % rule:: completed_TGS(a, ak(c,_), _)  -> sec.
    % rule:: completed_C_tgs(a, ak(c,_), enc(kS, _), sk(c))  -> sec.
    % rule:: request(a) -> sec.  
    % rule:: SK, completed_S(a, SK, _) -> sec.
    % rule:: Req, completed_S(a, _, Req) -> sec.
    % rule:: AK, completed_C(a, AK, _) -> sec.
    % rule:: SK, completed_C(a, _, SK) -> sec.
