%%
%%  SSH-Trasport
%%  
%%  Modelling follows the one from the AVISPA Library of Protocols.
%%
%%  The server uses the same exponent (Y) in all sessions.
%%


% Functors:

    % cryptographic primitives
    fun sig/2.
    fun enc/2.
    fun h/1.

    % participant names 
    fun a/0.
    fun b/0.
    fun c/0.

    % constants
    fun g/0.
    fun sec/0.
    fun letterC/0.
    fun letterD/0.

    % nonces and keys
    fun k/1.    % k(K) -- private key of X.
    fun pub/1.

    fun nc/2.  % ...(X,Y) -- a nonce generated by X in the session with Y.
    fun ns/2.  
    fun dhx/2. 
    fun dhy/1.
    fun secC/2.
    fun secS/1.

    fun completedC/5.
    fun completedS/5.


%%%%% AUXILIARY PREDICATES %%%%%%%%%%%%

participant(a).
participant(b).
participant(c).

%%
% The following predicates specify which session will be
% taken into consideration.

% roleC(X,Y)  ---  X as a client runs the protocol with Y as a server 
% roleS(X)    ---  X runs the protocol as a server 

roleC(a, b).
roleC(a, c).

roleS(b).

% roleC(C, S, NC, X, SecC)
%
roleC(C, S, nc(C,S), dhx(C,S), secC(C,S)) :- roleC(C,S).
    
% roleS(S, C, Y, SecS)
%
roleS(S, dhy(S), secS(S)) :- roleS(S).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The secret:

    query sec.

% Intruder rules

    rule:: sig(_,M) -> M.
    rule:: sig(K,_) -> pub(K).
    rule:: K, M -> sig(K,M).
    rule:: K, M -> enc(K,M).
    rule:: K, enc(K,M) -> M.
    rule:: M -> h(M).

% Initial intruder knowledge

    rule:: X             :- participant(X).
    rule:: pub(k(X))     :- participant(X).
    rule:: k(c).
    rule:: g.
    rule:: letterC.
    rule:: letterD.

% Protocol Rules (the main session):


    % (1) C->S
    %
    rule::
        NC
    :-
        roleC(_C, _S, NC, _X, _SecC).


    % (2) S->C
    %
    rule::
        NC -> ns(S,NC)
    :-
        roleS(S, _Y, _).


    % (3) C->S
    %
    rule::
        _NS  ->  g^X
    :-
        roleC(_C, _S, _NC, X, _SecC).


    % (4) S->C
    %
    rule::
        NC, % from the previous step
        GX
      ->
        pub(k(S)), g^Y, sig(k(S), H)
    :-
        roleS(S, Y, _),
        H = h(( NC, ns(S,NC), pub(k(S)), GX, g^Y, GX^Y )).


    % (5) C->S
    %
    rule::
        NS, % from previous steps
        pub(k(S)), GY, sig(k(S), H)
      ->
        enc(KCS, SecC)
    :-
        roleC(_C, S, NC, X, SecC),
        H = h(( NC, NS, pub(k(S)), g^X, GY, GY^X )),
        KCS = h(( GY^X, H, letterC, H )).


    % (6) S->C
    %
    rule::
        NC, GX, % from previous steps
        enc(KCS, _SecC)
      ->
        enc(KSC, SecS),
        completedS(S,GX,GX^Y,KCS,KSC)
    :-
        roleS(S, Y, SecS),
        H = h(( NC, ns(S,NC), pub(k(S)), GX, g^Y, GX^Y )),
        KCS = h(( GX^Y, H, letterC, H )),
        KSC = h(( GX^Y, H, letterD, H )).


    % (7) C completed
    %
    rule::
        NS, pub(k(S)), GY, sig(k(S), H), % from previous steps
        enc(KSC, _SecS)
      ->
        enc(KCS, SecC),
        completedC(C,S,GY^X,KCS,KSC)
    :-
        roleC(C, S, NC, X, SecC),
        H = h(( NC, NS, pub(k(S)), g^X, GY, GY^X )),
        KCS = h(( GY^X, H, letterC, H )),
        KSC = h(( GY^X, H, letterD, H )).


%%%%%%%%%%%%%%%%%%%%


    %%%
    %% should not fire
    %

    rule:: secC(a,b) -> sec.
    rule:: K, completedC(a,b,K,_,_) -> sec.
    rule:: KCS, completedC(a,b,_,KCS,_) -> sec.
    rule:: KSC, completedC(a,b,_,_,KSC) -> sec.
    rule:: K, completedS(b, g^dhx(a,c), _, _, K) -> sec.


    %%%
    %% should fire
    %
    
    % rule:: completedC(a,b,K,KCS,KSC), completedS(b,_,K,KCS,KSC) -> sec.
    % rule:: completedC(a,b,K,_,_), completedS(b,g^dhx(a,b),K,_,_) -> sec.
    % rule:: K, KCS, KSC, completedC(a,c,K,KCS,KSC)  -> sec.
    % rule:: K, KCS, KSC, completedS(b,_,K,KCS,KSC)  -> sec.
    
