%%
%%  STS with CA related attack.
%%  
%%  We model here the attack that is possible to run when parties can
%%  register public keys without proving knowledge of the
%%  corresponding private key.
%%
%%  The CA authority is modelled by fact of the form ca(X,PKX), where
%%  X is a participant name and PKX is its registered public key.


% Functors:

    % cryptographic primitives
    fun sig/2.
    fun enc/2.
    fun h/2.

    % CA
    fun ca/2.

    % participant names 
    fun a/0.
    fun b/0.
    fun c/0.

    % constants
    fun g/0.
    fun sec/0.

    % nonces and keys
    fun rI/2.   % rI(X,Y) -- a nonce (exponent) generated by X as an initiator for the session with Y.
    fun rR/2.   % rR(X,Y) -- a nonce (exponent) generated by X as a responder for the session with Y.
    fun k/1.    % k(K) -- private key of X.
    fun pub/1.
    fun key/2.  % key(X,Y) -- a key shared by X and Y.

    fun completedI/3.
    fun completedR/3.


%%%%% AUXILIARY PREDICATES %%%%%%%%%%%%

participant(a).
% participant(b). 
participant(c). % dishonest

honest(a).

%%
% The following predicates specify which session will be
% taken into consideration.

% roleI(X,Y)  ---  X as an initiator runs the protocol with Y as responder 
% roleR(X,Y)  ---  X as a responder runs the protocol with Y as initiator 
%

roleI(a,a).
roleI(a,c).
roleR(a,a).
roleR(a,c).

%% roleI(a, b).
%% roleI(a, c).
%% roleI(b, a).
%% roleI(b, c).
%% 
%% roleR(a, b).
%% roleR(a, c).
%% roleR(b, a).
%% roleR(b, c).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% The secret:

    query sec.

% Intruder rules

    rule:: sig(_,M) -> M.
    rule:: K, M -> sig(K,M).
    rule:: K, M -> enc(K,M).
    rule:: K, enc(K,M) -> M.
    rule:: K, M -> h(K,M).

% Modelling CA

    % Honest participants have regitered their public keys properly:
    rule:: ca(A,pub(k(A)))  :- honest(A).

    % A dishonest participant can register anything as his public key:
    rule:: X -> ca(c,X).

% Initial intruder knowledge

    rule:: X             :- participant(X).
    rule:: pub(k(X))     :- participant(X).
    rule:: k(c).
    rule:: rI(c,X)       :- participant(X).
    rule:: rR(c,X)       :- participant(X).
    rule:: key(c,X)      :- participant(X).
    rule:: key(X,c)      :- participant(X).
    rule:: g.

    % rule k(a).

% Protocol Rules (the main session):

    % A->B:
    rule::  
        g^rI(A,B)
    :- 
        roleI(A,B).

    % B->A:
    rule::  
        GA ->  g^rR(B,A), B, enc(GA^rR(B,A), sig(k(B),(GA,g^rR(B,A))))    
    :- 
        roleR(B,A).

    % A->B:
    rule::  
        GB, B, ca(B,pub(KB)), enc(GB^rI(A,B), sig(KB,(g^rI(A,B),GB)))  
            ->  A, enc(GB^rI(A,B), sig(k(A),(GB,g^rI(A,B)))),  completedI(A,B,GB^rI(A,B))
    :- 
        roleI(A,B).

    % B comleted:
    rule:: 
        A, ca(A,pub(KA)), enc(GA^rR(B,A), sig(KA, (g^rR(B,A),GA))) -> completedR(B,A,GA^rR(B,A))
    :- 
        roleR(B,A).

%%%%%%
    
    %%%
    %% should not fire
    %

    rule:: K, completedI(a,a,K) -> sec.
    rule:: K, completedR(a,a,K) -> sec.

    % rule:: completedI(a,a,K), completedR(a,c,K) -> sec.
    rule:: completedI(a,c,K), completedR(a,a,K) -> sec.

    rule:: g^rI(a,a)^rR(a,a) -> sec.
    
